"""Cell template class"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


# pylint: disable=W0511

# TODO take into account that one might want to run protocols on different
# machines
# TODO rename this to 'CellModel' -> definitely

import collections
import os

from bluepyopt.ephys import morphologies

import logging
logger = logging.getLogger(__name__)


class Model(object):

    """Model"""

    def __init__(self, name):
        """Constructor
        Args:
            name (str): name of the model
        """
        self.name = name

    def instantiate(self, sim=None):
        """Instantiate model in simulator"""
        pass

    def destroy(self, sim=None):
        """Destroy instantiated model in simulator"""
        pass


class CellModel(Model):

    """Cell model class"""

    def __init__(
            self,
            name,
            morph=None,
            mechs=None,
            params=None):
        """Constructor

        Args:
            name (str): name of this object
            morph (Morphology):
                underlying Morphology of the cell
            mechs (list of Mechanisms):
                Mechanisms associated with the cell
            params (list of Parameters):
                Parameters of the cell model
        """
        super(CellModel, self).__init__(name)
        self.morphology = morph
        self.mechanisms = mechs
        self.params = collections.OrderedDict()
        for param in params:
            self.params[param.name] = param

        # Cell instantiation in simulator
        self.icell = None

        self.param_values = None

    def params_by_names(self, param_names):
        """Get parameter objects by name"""

        return [self.params[param_name] for param_name in param_names]

    def freeze(self, param_dict):
        """Set params"""

        for param_name, param_value in param_dict.items():
            self.params[param_name].freeze(param_value)

    def unfreeze(self, param_names):
        """Unset params"""

        for param_name in param_names:
            self.params[param_name].unfreeze()

    @staticmethod
    def create_empty_template(template_name):
        '''create an hoc template named template_name for an empty cell'''
        return '''\
        begintemplate %(template_name)s
          objref all, apical, basal, somatic, axonal, this, CellRef
          proc init() {
            all = new SectionList()
            somatic = new SectionList()
            basal = new SectionList()
            apical = new SectionList()
            axonal = new SectionList()
            forall delete_section()
            CellRef = this
          }

          proc destroy() {localobj nil
            CellRef = nil
          }

          create soma[1], dend[1], apic[1], axon[1]
        endtemplate %(template_name)s
               ''' % dict(template_name=template_name)

    @staticmethod
    def create_empty_cell(name, sim):
        """Create an empty cell in Neuron"""

        # TODO minize hardcoded definition
        # E.g. sectionlist can be procedurally generated
        hoc_template = CellModel.create_empty_template(name)
        sim.neuron.h(hoc_template)

        template_function = getattr(sim.neuron.h, name)

        return template_function()

    def instantiate(self, sim=None):
        """Instantiate model in simulator"""

        # TODO replace this with the real template name
        if not hasattr(sim.neuron.h, 'Cell'):
            self.icell = self.create_empty_cell('Cell', sim=sim)
        else:
            self.icell = sim.neuron.h.Cell()

        self.morphology.instantiate(sim=sim, icell=self.icell)

        for mechanism in self.mechanisms:
            mechanism.instantiate(sim=sim, icell=self.icell)
        for param in self.params.values():
            param.instantiate(sim=sim, icell=self.icell)

    def destroy(self, sim=None):  # pylint: disable=W0613
        """Destroy instantiated model in simulator"""

        # Make sure the icell's destroy() method is called
        # without it a circular reference exists between CellRef and the object
        # this prevents the icells from being garbage collected, and
        # cell objects pile up in the simulator
        self.icell.destroy()

        # The line below is some M. Hines magic
        # DON'T remove it, because it will make sure garbage collection
        # is called on the icell object
        sim.neuron.h.Vector().size()

        self.icell = None

        self.morphology.destroy(sim=sim)
        for mechanism in self.mechanisms:
            mechanism.destroy(sim=sim)
        for param in self.params.values():
            param.destroy(sim=sim)

    def check_nonfrozen_params(self, param_names):  # pylint: disable=W0613
        """Check if all nonfrozen params are set"""

        for param_name, param in self.params.items():
            if not param.frozen:
                raise Exception(
                    'CellModel: Nonfrozen param %s needs to be '
                    'set before simulation' %
                    param_name)

    def create_hoc(self, param_values, template_name='CCell',
                   ignored_globals=(), template='cell_template.jinja2'):
        from bluepyopt.ephys.create_hoc import create_hoc

        to_unfreeze = []
        for param in self.params.values():
            if not param.frozen:
                param.freeze(param_values[param.name])
                to_unfreeze.append(param.name)

        ret = create_hoc(mechanisms=self.mechanisms,
                         parameters=self.params.values(),
                         morphology=self.morphology.morphology_path,
                         ignored_globals=ignored_globals,
                         template_name=template_name,
                         template=template)

        self.unfreeze(to_unfreeze)

        return ret


    def __str__(self):
        """Return string representation"""

        content = '%s:\n' % self.name

        content += '  morphology:\n'
        content += '    %s\n' % str(self.morphology)

        content += '  mechanisms:\n'
        for mechanism in self.mechanisms:
            content += '    %s\n' % mechanism
        content += '  params:\n'
        for param in self.params.values():
            content += '    %s\n' % param

        return content


def load_hoc_template(sim, hoc_path):
    '''have neuron load a hoc file, and detect what the name template name is

    Note: this may fail if there is a begintemplate in a /* */ style comment

    The template must have an init that takes two parameters, the second of
    which is the path to a morphology.

    It must also have a CellRef member that is the result of
        `Import3d_GUI(...).instantiate()`
    '''
    with open(hoc_path) as fd:
        for i, line in enumerate(fd):
            if 'begintemplate' in line:
                line = line.strip().split()
                assert line[0] == 'begintemplate', \
                    'begintemplate must come first, line %d' % i
                template_name = line[1]
                logger.info('Found template %s on line %d', template_name, i)
                break
        else:
            raise Exception('Could not find begintemplate in hoc file')

    if not hasattr(sim.neuron.h, template_name):
        sim.neuron.h.load_file(hoc_path)
        assert hasattr(sim.neuron.h, template_name), \
            'NEURON does not have template: ' + template_name

    return template_name


class HocMorphology(morphologies.Morphology):

    '''wrapper for Morphology so that it has a morphology_path'''

    def __init__(self, morphology_path):
        if not os.path.exists(morphology_path):
            raise Exception('HocCellModel: Morphology not found at: %s'
                            % morphology_path)
        self.morphology_path = morphology_path

class HocCellModel(CellModel):

    '''Wrapper class for a hoc template so it can be used by BluePyOpt'''
    def __init__(self, name, morphology_path, hoc_path):
        """Constructor

        Args:
            name(str): name of this object
            sim(NrnSimulator): simulator in which to instatiate hoc_path
            morphology_path(str path): path to morphology that can be loaded by
                                       Neuron
            hoc_path(str path): path to .hoc file that will be used
        """
        super(HocCellModel, self).__init__(name,
                                           morph=None,
                                           mechs=[],
                                           params=[])
        self.hoc_path = hoc_path
        self.morphology = HocMorphology(morphology_path)
        self.cell = None
        self.icell = None

    def params_by_names(self, param_names):
        pass

    def freeze(self, param_dict):
        pass

    def unfreeze(self, param_names):
        pass

    def instantiate(self, sim=None):
        sim.neuron.h.load_file('stdrun.hoc')
        template_name = load_hoc_template(sim, self.hoc_path)
        morph_path = self.morphology.morphology_path
        self.cell = getattr(sim.neuron.h, template_name)(0, morph_path)
        self.icell = self.cell.CellRef

    def destroy(self, sim=None):
        self.cell = None
        self.icell = None

    def check_nonfrozen_params(self, param_names):
        pass

    def __str__(self):
        """Return string representation"""
        return ('%s: %s of %s(%s)' %
                (self.__class__, self.name, self.hoc_path,
                 self.morphology.morphology_path, ))



"""Run simple cell optimisation"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
# pylint: disable=R0914

import os
import json

import bluepyopt.ephys as ephys

script_dir = os.path.dirname(__file__)
config_dir = os.path.join(script_dir, 'config')

# TODO store definition dicts in json
# TODO rename 'score' into 'objective'
# TODO add functionality to read settings of every object from config format


def define_mechanisms():
    """Define mechanisms"""

    mech_definitions = json.load(
        open(
            os.path.join(
                config_dir,
                'mechanisms.json')))

    mechanisms = []
    for sectionlist, channels in mech_definitions.iteritems():
        seclist_loc = ephys.locations.NrnSeclistLocation(
            sectionlist,
            seclist_name=sectionlist)
        for channel in channels:
            mechanisms.append(ephys.mechanisms.NrnMODMechanism(
                name='%s.%s' % (channel, sectionlist),
                mod_path=None,
                prefix=channel,
                locations=[seclist_loc],
                preloaded=True))

    return mechanisms


def define_parameters():
    """Define parameters"""

    param_configs = json.load(open(os.path.join(config_dir, 'parameters.json')))
    parameters = []

    for param_config in param_configs:
        if 'value' in param_config:
            frozen = True
            value = param_config['value']
            bounds = None
        elif 'bounds':
            frozen = False
            bounds = param_config['bounds']
            value = None
        else:
            raise Exception(
                'Parameter config has to have bounds or value: %s'
                % param_config)

        if param_config['type'] == 'global':
            parameters.append(
                ephys.parameters.NrnGlobalParameter(
                    name=param_config['param_name'],
                    param_name=param_config['param_name'],
                    frozen=frozen,
                    bounds=bounds,
                    value=value))
        elif param_config['type'] in ['section', 'range']:
            if param_config['dist_type'] == 'uniform':
                scaler = ephys.parameterscalers.NrnSegmentLinearScaler()
            elif param_config['dist_type'] == 'exp':
                scaler = ephys.parameterscalers.NrnSegmentSomaDistanceScaler(
                    distribution=param_config['dist'])
            seclist_loc = ephys.locations.NrnSeclistLocation(
                param_config['sectionlist'],
                seclist_name=param_config['sectionlist'])

            name = '%s.%s' % (param_config['param_name'],
                              param_config['sectionlist'])

            if param_config['type'] == 'section':
                parameters.append(
                    ephys.parameters.NrnSectionParameter(
                        name=name,
                        param_name=param_config['param_name'],
                        value_scaler=scaler,
                        value=value,
                        frozen=frozen,
                        bounds=bounds,
                        locations=[seclist_loc]))
            elif param_config['type'] == 'range':
                parameters.append(
                    ephys.parameters.NrnRangeParameter(
                        name=name,
                        param_name=param_config['param_name'],
                        value_scaler=scaler,
                        value=value,
                        frozen=frozen,
                        bounds=bounds,
                        locations=[seclist_loc]))
        else:
            raise Exception(
                'Param config type has to be global, section or range: %s' %
                param_config)

    return parameters


def define_morphology():
    """Define morphology"""

    return ephys.morphologies.NrnFileMorphology(
        os.path.join(
            script_dir,
            'morphology/C060114A7.asc'),
        do_replace_axon=True)


def create():
    """Create cell model"""

    cell = ephys.models.CellModel(
        'l5pc',
        morph=define_morphology(),
        mechs=define_mechanisms(),
        params=define_parameters())

    return cell



#!/usr/bin/env python

'''Example for generating a hoc template

 $ python generate_hoc.py > test.hoc

 Will save 'test.hoc' file, which can be loaded in neuron with:
     'load_file("test.hoc")'
 Then the hoc template needs to be instantiated with a morphology
     CCell("ignored", "path/to/morphology.swc")
'''
import sys

from bluepyopt.ephys.models import CellModel
import l5pc_model

def main():
    '''main'''
    param_values = {
        'gNaTs2_tbar_NaTs2_t.apical': 0.026145,
        'gSKv3_1bar_SKv3_1.apical': 0.004226,
        'gImbar_Im.apical': 0.000143,
        'gNaTa_tbar_NaTa_t.axonal': 3.137968,
        'gK_Tstbar_K_Tst.axonal': 0.089259,
        'gamma_CaDynamics_E2.axonal': 0.002910,
        'gNap_Et2bar_Nap_Et2.axonal': 0.006827,
        'gSK_E2bar_SK_E2.axonal': 0.007104,
        'gCa_HVAbar_Ca_HVA.axonal': 0.000990,
        'gK_Pstbar_K_Pst.axonal': 0.973538,
        'gSKv3_1bar_SKv3_1.axonal': 1.021945,
        'decay_CaDynamics_E2.axonal': 287.198731,
        'gCa_LVAstbar_Ca_LVAst.axonal': 0.008752,
        'gamma_CaDynamics_E2.somatic': 0.000609,
        'gSKv3_1bar_SKv3_1.somatic': 0.303472,
        'gSK_E2bar_SK_E2.somatic': 0.008407,
        'gCa_HVAbar_Ca_HVA.somatic': 0.000994,
        'gNaTs2_tbar_NaTs2_t.somatic': 0.983955,
        'decay_CaDynamics_E2.somatic': 210.485284,
        'gCa_LVAstbar_Ca_LVAst.somatic': 0.000333,
    }
    cell = l5pc_model.create()
    print cell.create_hoc(param_values)

if __name__ == '__main__':
    if '-h' in sys.argv or '--help' in sys.argv:
        print __doc__
    else:
        main()



"""Create tables for BluePyOpt paper"""

import json


def load_json(filename):
    """Load struct from json"""

    return json.load(open(filename))


def create_feature_fields():
    """Create fields for param.json"""

    features = json.load(open('features.json'))

    fields_content = ''

    print features

    fields_content += \
        '\t\tStimulus & Location & eFeature & Mean & Std \\\\ \n'
    fields_content += '\t\t\\midrule\n'
    for stimulus, loc_list in sorted(features.iteritems()):
        stim_field = stimulus
        for location, features in sorted(loc_list.iteritems()):
            loc_field = location
            for feature_name, (mean, std) in sorted(features.iteritems()):
                feature_name = feature_name.replace('_', '{\\_}')
                fields_content += '\t\t%s \\\\\n' % ' & '.join([stim_field,
                                                                loc_field,
                                                                feature_name,
                                                                str(mean),
                                                                str(std)])
                if loc_field is not '':
                    loc_field = ''
                if stim_field is not '':
                    stim_field = ''
    fields_content += '\t\t\\botrule\n'

    return fields_content, 6


def create_param_fields_list():
    """Create fields for param.json"""

    import collections
    param_configs = json.load(open('parameters.json'))

    fields_list = []
    for param_config in param_configs:
        fields = collections.OrderedDict(
            (('location', ''),
             ('mech', ''),
             ('param', ''),
             ('dist', ''),
             ('units', ''),
             ('lbound', ''),
             ('ubound', ''),
             ('value', '')))

        if 'value' in param_config:
            fields['value'] = str(param_config['value'])
        elif 'bounds' in param_config:
            fields['lbound'] = str(param_config['bounds'][0])
            fields['ubound'] = str(param_config['bounds'][1])

        if param_config['type'] == 'global':
            fields['param'] = param_config['param_name']
            fields['location'] = 'global'
        elif param_config['type'] in ['section', 'range']:
            fields['dist'] = param_config['dist_type']
            fields['location'] = param_config['sectionlist']
            if param_config['type'] == 'range':
                fields['mech'] = param_config['mech']
                fields['param'] = param_config['mech_param']
            elif param_config['type'] == 'section':
                fields['param'] = param_config['param_name']

        if 'bar' in fields['param'] or fields['param'] == 'g_pas':
            fields['units'] = '\\SI{}{\\siemens\\per\\cm\\squared}'
        elif fields['param'] == 'gamma':
            fields['units'] = ''
        elif fields['param'] == 'decay':
            fields['units'] = ' \\SI{}{\\milli\\second}'
        elif 'celsius' == fields['param']:
            fields['units'] = ' \\SI{}{\\celsius}'
        elif fields['param'] in ['v_init', 'e_pas', 'ek', 'ena']:
            fields['units'] = ' \\SI{}{\\milli\\volt}'
        elif 'Ra' == fields['param']:
            fields['units'] = ' \\SI{}{\\ohm\\cm}'
        elif 'cm' == fields['param']:
            fields['units'] = ' \\SI{}{\\micro\\farad\\per\\cm\\squared}'

        for key in fields:
            fields[key] = fields[key].replace('_', '{\\_}')

        fields_list.append(fields)

    return fields_list


def create_param_fields_string():
    """Create parameter fields string"""

    fields_list = create_param_fields_list()

    fields_content = ''
    fields_content += \
        '\t\tLocation & Mechanism & Parameter name & ' \
        'Distribution & Units & Lower bound & Upper bound \\\\\n'
    fields_content += '\t\t\\midrule\n'

    opt_fields = [field for field in fields_list if field['value'] == '']
    fixed_fields = [field for field in fields_list if field['value'] != '']
    global_fixed_fields = [
        field
        for field in fixed_fields if field['location'] == 'global']
    nonglobal_fixed_fields = [
        field
        for field in
        fixed_fields if field['location'] != 'global']

    for fields in opt_fields:
        del fields['value']
        fields_content += '\t\t%s \\\\\n' % ' & '.join(fields.values())
    fields_content += '\t\t\\midrule\n'
    fields_content += \
        '\t\tLocation & Mechanism & Parameter name & ' \
        'Distribution & Units & Value & \\\\\n'
    fields_content += '\t\t\\midrule\n'
    for fields in global_fixed_fields + nonglobal_fixed_fields:
        del fields['lbound']
        del fields['ubound']
        fields_content += '\t\t%s \\\\\n' % ' & '.join(
            fields.values() + [' '])

    fields_content += '\t\t\\botrule\n'

    return fields_content, 7


def create_table(field_content, n_of_cols):
    """Surround fiels with table creation"""

    table_content = '\\begin{tabular}{%s}\n\t\\toprule\n' % \
        ('l' * n_of_cols)

    table_content += field_content

    table_content += '\n\\end{tabular}'

    return table_content


def main():
    """Main"""

    param_fields, n_of_cols = create_param_fields_string()
    param_content = create_table(param_fields, n_of_cols)
    open('tables/params.tex', 'w').write(param_content)
    print param_content

    feature_fields, n_of_cols = create_feature_fields()
    feature_content = create_table(feature_fields, n_of_cols)
    open('tables/features.tex', 'w').write(feature_content)
    print feature_content

if __name__ == '__main__':
    main()



#!/usr/bin/env python
'''Utility functions to read the tasks.db from ipyparallel when used
with BluePyOpt.

The script is executable, as an example on how to extract the top 10
longest running engines for the l5pc model.
'''

import os
import sqlite3

import pandas
import numpy as np

from ipyparallel.controller import sqlitedb


def _open_db(dbfile):
    '''open the sqlite file, and set up the converters used to serialize data'''
    sqlite3.register_adapter(dict, sqlitedb._adapt_dict)
    sqlite3.register_converter('dict', sqlitedb._convert_dict)
    sqlite3.register_adapter(list, sqlitedb._adapt_bufs)
    sqlite3.register_converter('bufs', sqlitedb._convert_bufs)
    db = sqlite3.connect(dbfile,
                         detect_types=sqlite3.PARSE_DECLTYPES,
                         cached_statements=64)
    return db


def _add_buffers(df, arg_names=None, result_names=None, delete_buffers=False):
    '''add argument and result buffers to the dataframe'''
    if arg_names is None:
        row = df['buffers'][0]
        count = len(sqlitedb._convert_bufs(row[3])[0])
        arg_names = ['arg_%02d' % i for i in range(count)]

    args = np.empty(shape=(df.shape[0], len(arg_names)))
    for i, row in enumerate(df['buffers']):
        # row[0] is pickled ref to __builtin__.map,
        # row[1] is pickled description of args:
        #  {'kw_keys': [], 'nargs': 2, 'narg_bufs': 2}
        # row[2] is pickled function to be mapped
        # row[3] are the pickled arguments, packed into a list
        args[i, :] = sqlitedb._convert_bufs(row[3])[0]
    args_df = pandas.DataFrame(args, columns=arg_names)

    if result_names is None:
        row = df['result_buffers'][0]
        count = len(sqlitedb._convert_bufs(row[0])[0])
        result_names = ['res_%02d' % i for i in range(count)]

    results = np.empty(shape=(df.shape[0], len(result_names)))
    for i, row in enumerate(df['result_buffers']):
        if row:
            # row[0] are the pickled results, in a list
            results[i, :] = sqlitedb._convert_bufs(row[0])[0]
        else:
            # either the task didn't return, or returned an error
            results[i, :] = float('NaN')
    results_df = pandas.DataFrame(results, columns=result_names)

    if delete_buffers:
        df.drop(['buffers', 'result_buffers'], axis=1, inplace=True)

    return pandas.concat([df, args_df, results_df], axis=1, join_axes=[df.index])


def create_df(db, arg_names=None, result_names=None):
    '''convert a tasks_db to a pandas dataframe'''
    df = pandas.read_sql('select * from `ipython-tasks`;', db,
                         parse_dates=('submitted', 'started', 'completed',))
    df = _add_buffers(df, arg_names, result_names)
    return df


if __name__ == '__main__':
    import sys

    from l5pc_model import define_parameters
    arg_names = [p.name for p in define_parameters() if not p.frozen]

    from l5pc_evaluator import (define_protocols, define_fitness_calculator)
    fitness_protocols = define_protocols()
    fitness_calculator = define_fitness_calculator(fitness_protocols)
    result_names = [o.name for o in fitness_calculator.objectives]

    db_filename = sys.argv[1]
    db = _open_db(db_filename)
    df = create_df(db, arg_names, result_names)
    df['run_time'] = df['completed'] - df['started']

    pandas.set_option('display.expand_frame_repr', False)
    title = '10 Longest times'
    print '%s\n%s' % (title, '*' * len(title))
    print df.nlargest(10, 'run_time')[['run_time', ] + arg_names + result_names]



"""Read exp data for validation"""

import igorpy
import numpy

voltages = []

# import matplotlib.pyplot as plt
# fig, axes = plt.subplots(11)

for i in range(436, 447):
    # header, voltage = igorpy.read('exp_data/X_NoiseSpiking_ch1_%d.ibw' % i)
    header, current = igorpy.read('exp_data/X_NoiseSpiking_ch0_%d.ibw' % i)
    # axes[i - 436].plot(voltage)
# header, voltage = igorpy.read('exp_data/X_APThreshold_ch1_262.ibw')
# header, current = igorpy.read('exp_data/X_APThreshold_ch0_262.ibw')
# plt.show()
# voltage *= 1000
# voltage -= 14.0
current *= 1e9

time = numpy.arange(len(current)) * header.dx * 1000

numpy.savetxt('exp_data/noise_i.txt', numpy.vstack((time, current)).T)

# import matplotlib.pyplot as plt

# fig, ax = plt.subplots(2)
# ax[0].plot(time, current)
# ax[1].plot(time, voltage)

# plt.show()



'''create a hoc file from a set of BluePyOpt.ephys parameters'''
import os
import re

from collections import defaultdict, namedtuple, OrderedDict

import jinja2
from bluepyopt.ephys.parameters import (NrnGlobalParameter,
                                        NrnSectionParameter,
                                        NrnRangeParameter,
                                        )
from bluepyopt.ephys.parameterscalers import (NrnSegmentSomaDistanceScaler,
                                              NrnSegmentLinearScaler,
                                              FLOAT_FORMAT,
                                              format_float,
                                              )

Location = namedtuple('Location', 'name, value')
Range = namedtuple('Range', 'location, param_name, value')
LOCATION_ORDER = ('all', 'apical', 'axonal', 'basal', 'somatic',
                  )


def _generate_channels_by_location(mechanisms):
    '''create a OrderedDictionary of all channel mechanisms for the hoc template'''
    channels = OrderedDict((location, []) for location in LOCATION_ORDER)
    for mech in mechanisms:
        name = mech.prefix
        for location in mech.locations:
            channels[location.seclist_name].append(name)
    return channels


def _generate_parameters(parameters):
    '''create a list of parameters that need to be added to the hoc template'''
    param_locations = defaultdict(list)
    global_params = {}
    for param in parameters:
        if isinstance(param, NrnGlobalParameter):
            global_params[param.name] = param.value
        else:
            assert isinstance(param.locations, (tuple, list)), 'Must have locations list'
            for location in param.locations:
                param_locations[location.seclist_name].append(param)

    section_params = defaultdict(list)
    range_params = []
    for loc in LOCATION_ORDER:
        if loc not in param_locations:
            continue
        for param in param_locations[loc]:
            if isinstance(param, NrnRangeParameter):
                if isinstance(param.value_scaler, NrnSegmentSomaDistanceScaler):
                    value = param.value_scaler.distribution
                    value = re.sub(r'math\.', '', value)
                    value = re.sub('{distance}', FLOAT_FORMAT, value)
                    value = re.sub('{value}', format_float(param.value), value)
                    range_params.append(Range(loc, param.param_name, value))
                elif isinstance(param.value_scaler, NrnSegmentLinearScaler):
                    value = param.value_scale_func(param.value)
                    section_params[loc].append(
                        Location(param.param_name, format_float(value)))
            elif isinstance(param, NrnSectionParameter):
                value = param.value_scale_func(param.value)
                section_params[loc].append(
                    Location(param.param_name, format_float(value)))

    ordered_section_params = [(loc, section_params[loc])
                              for loc in LOCATION_ORDER]

    return global_params, ordered_section_params, range_params


def create_hoc(mechanisms, parameters, morphology=None, ignored_globals=(),
               template_name='CCell', template='cell_template.jinja2'):
    '''return a string containing the hoc template

    Args:
        mechanisms(): All the mechanisms for the hoc template
        parameters(): All the parameters in the hoc template
        morpholgy(str path): Path to morphology
        ignored_globals(iterable str): HOC coded is added for each NrnGlobalParameter
        that exists, to test that it matches the values set in the parameters.  This
        iterable contains parameter names that aren't checked
        template(str): name of the template to use 'cell_template.jinja2',
    '''
    templates_basepath = os.path.abspath(os.path.dirname(__file__))
    template = os.path.join(templates_basepath, 'templates', template)
    with open(template) as fd:
        template = fd.read()
        template = jinja2.Template(template)

    channels = _generate_channels_by_location(mechanisms)
    global_params, section_params, range_params = \
        _generate_parameters(parameters)

    ignored_global_params = {}
    for ignored_global in ignored_globals:
        if ignored_global in global_params:
            ignored_global_params[ignored_global] = global_params[ignored_global]
            del global_params[ignored_global]

    return template.render(template_name=template_name,
                           channels=channels,
                           morphology=morphology,
                           section_params=section_params,
                           range_params=range_params,
                           global_params=global_params,
                           ignored_global_params=ignored_global_params)



"""Convert params.json and fixed_params.json to parameters.json format"""

from __future__ import print_function

import json


def main():
    """Main"""

    fixed_params = json.load(open('fixed_params.json'))
    params = json.load(open('params.json'))

    parameters = []

    for sectionlist in fixed_params:
        if sectionlist == 'global':
            for param_name, value in fixed_params[sectionlist]:
                param = {
                    'value': value,
                    'param_name': param_name,
                    'type': 'global'}
                parameters.append(param)
        else:
            for param_name, value, dist_type in fixed_params[sectionlist]:
                param = {
                    'value': value,
                    'param_name': param_name,
                    'type': 'section',
                    'dist_type': dist_type,
                    'sectionlist': sectionlist
                }
                parameters.append(param)

    for sectionlist in params:
        for mech, param_name, min_bound, max_bound, dist_type in \
                params[sectionlist]:
            param = {
                'bounds': [min_bound, max_bound],
                'mech': mech,
                'mech_param': param_name,
                'param_name': '%s_%s' % (param_name, mech),
                'type': 'range',
                'dist_type': dist_type,
                'sectionlist': sectionlist
            }

            if mech == 'Ih':
                del param['bounds']
                param['value'] = 8e-5

            if dist_type == 'exp':
                param['dist'] = \
                    '(-0.8696 + 2.087*math.exp(({distance})*0.0031))*{value}'

            parameters.append(param)

    json.dump(parameters, open('parameters.json', 'w'),
              indent=4,
              separators=(',', ': '))

if __name__ == '__main__':
    main()



"""Init script"""

"""
Copyright (c) 2016, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=W0511

import simulators  # NOQA
import models  # NOQA
import evaluators  # NOQA
import mechanisms  # NOQA
import locations  # NOQA
import parameterscalers  # NOQA
import parameters  # NOQA
import morphologies  # NOQA
import efeatures  # NOQA
import objectives  # NOQA
import protocols  # NOQA
import responses  # NOQA
import recordings  # NOQA
import objectivescalculators  # NOQA
import stimuli  # NOQA

# TODO create all the necessary abstract methods
# TODO check inheritance structure
# TODO instantiate using 'simulation env' as parameter, instead of cell
